local Goranaws_AuraTemplate = {}
Goranaws_TemplateManager = {}

local Masque = LibStub('Masque', true)

function Goranaws_TemplateManager.Merge(source, destination)
	if not destination then
		destination = {}
	end
	if not source then
		return destination
	end
	for i, b in pairs(source) do
		if (type(b) == 'table') then
			destination[i] = Merge(b, destination[i])
		else
			destination[i] = b
		end
	end
	return destination
end

function Goranaws_TemplateManager:New(self, kind)
	Goranaws_TemplateManager.Merge(Goranaws_AuraTemplate, self)
	self.kind = kind
	if Masque then
		self.masque = self.masque or Masque:Group("Dominos", kind)
	end
	self:GetOrCreateOverlays()
	self:CreateHeader(kind)

	self:Layout()
	self:UpdateFilter()
	self:UpdateTarget()
	self:UpdateAuras()
end

local temp = Goranaws_AuraTemplate

function temp:NumButtons()
	return self:NumColumns() * self:NumRows()
end

function temp:NumColumns()
	return self.sets.columns
end

function temp:NumRows()
	return self.sets.rows
end

function temp:UpdateAuraLayout(spacing, cols, rows, LR, TB, method, direction)
	for i,aura in self:GetParent():ActiveChildren() do
		if (tonumber(cols)and tonumber(rows)) and (tonumber(i) > (tonumber(cols)* tonumber(rows))) then
			aura:Hide()
		end
	end
	local base = (30 + spacing)
	self:SetAttribute("minWidth", base);
	self:SetAttribute("minHeight", base);

	self:SetAttribute("wrapAfter", cols);
	self:SetAttribute("maxWraps", rows);
	
	local hori, vert
	
	if not LR then	
		vert = "Left"
		self:SetAttribute("xOffset", base);
		self:SetAttribute("yOffset", 0);
	else
		vert = "Right"
		self:SetAttribute("xOffset", -base);
		self:SetAttribute("yOffset", 0);
	end
	if not TB then
		hori = "Top"
		self:SetAttribute("wrapXOffset", 0);
		self:SetAttribute("wrapYOffset", -base);	
	else
		hori = "Bottom"
		self:SetAttribute("wrapXOffset", 0);
		self:SetAttribute("wrapYOffset", base);
	end
	self:SetAttribute("point", hori..vert);
	self:SetAttribute("sortMethod", method); -- INDEX or NAME or TIME
	self:SetAttribute("sortDirection", direction); -- - to reverse
end

function temp:CreateHeader()
	self.headerAura = self.headerAura or CreateFrame("Frame", "Goranaws_"..self.kind.."_Bar", self, "SecureAuraHeaderTemplate")
	self.total = self.total or 0
	RegisterUnitWatch(self.headerAura)
	
	self.headerAura:SetScript("OnShow", function()
		self:UpdateAuras()
	end)
	self.headerAura:SetScript("OnHide", function()
		self:ResetAuras()
	end)
	self:SetScript("OnEvent", function(s, event, ...)
		self:OnEvent(event, ...)
	end)
	
	self.headerAura:SetAttribute("template", "GoranawsAuraTemplate"); -- must be the template name of your XML
	
	local headerAura = self.headerAura

	self.headerAura.UpdateLayout = temp.UpdateAuraLayout
	self:RegisterEvent("UNIT_AURA");
	self:RegisterEvent("UNIT_TARGET");
	self:RegisterEvent("PLAYER_ENTERING_WORLD");	
	
	self.headerAura:UpdateLayout(3, "TopLeft", 4, 4, true, true,"Time", "+")
end

local function siter_active_children(header, i)
	i = i + 1;
	local child = header:GetAttribute("child" .. i);
	if child and child:IsShown() then
		return i, child, child:GetAttribute("index");
	end
end

function temp:ActiveChildren()
	return function(header, i)
		i = i + 1;
		local child = header:GetAttribute("child" .. i);
		if child and child:IsShown() then
			return i, child, child:GetAttribute("index");
		end
	end, self.headerAura, 0
end

function temp:GetOrCreateOverlays()
	self.overButtons = self.overButtons or {}
	local O = self.overButtons
	for i = 1, self:NumButtons() do
		O[i] = O[i] or CreateFrame("Button", "Goranaws_"..self.kind.."_OverlayButton_"..i, self, "GoranawsOverlayAuraTemplate")
		local b = O[i]
		b.cooldown:SetAllPoints(b)
		b:EnableMouse(false)
		if self.masque then
			self.masque:AddButton(b, {Icon = b.icon,Cooldown = b.cooldown})
		end
	end
end



function temp:UpdateOverlay(aura)
	local i = aura:GetID()
	if not  self.overButtons then
		self:GetOrCreateOverlays()
	end
	local b = self.overButtons[i]
	b:SetAllPoints(aura)
	local name, _, icon, count, debuffType, duration, expirationTime = UnitAura(self.headerAura:GetAttribute("unit"), i, self.headerAura:GetAttribute("filter"));
	if name then
		b.icon:SetTexture(icon);
		if _G[OmniCC] then
			b.cooldown:SetHideCountdownNumbers(false)
		else
			b.cooldown:SetHideCountdownNumbers(true)
		end
		b.cooldown:SetCooldown(expirationTime - duration, duration);
			b.cooldown:Show()


		if not (count>1)then count=""end
		b.txt:SetText(count);
		b:Show()
	else
		b.icon:SetTexture("");
		b.cooldown:Hide();
		b.txt:SetText("");
		b:Hide()
	end
end

function temp:UpdateAuras()
	self:ResetAuras()
	self.total = 0

	for i,aura in self:ActiveChildren() do
		self.total = self.total + 1
		
		local aura = self.headerAura:GetAttribute("child" .. i)
		self:UpdateOverlay(aura)
	end
end

function temp:ResetAuras()
	for i = 1, self.total do
		local aura = self.headerAura:GetAttribute("child" .. i)
		self:UpdateOverlay(aura)
	end
end

function temp:UpdateFilter()
	self.headerAura:SetAttribute("filter", self:GetFilter()); -- to activate UNITAURA event refresh
	self:GetScript("OnEvent")()
end

function temp:GetTarget()
	return self.sets.target or "player"
end

function temp:UpdateTarget()
	self.headerAura:SetAttribute("unit", self:GetTarget())
	self:GetScript("OnEvent")()
end

function temp:SetTarget(unit)
	self.sets.target = unit or "player"
	self:UpdateTarget()
end

function temp:GetSpacing()
	return self.sets.spacing
end

function temp:SetSpacing(spacing)
    self.sets.spacing = spacing
    self:Layout()
	
end

function temp:OnEvent(event, ...)
	self:UpdateAuras()
end

local items = {"Time", "Index", "Name"}

function temp:Layout()
	if not InCombatLockdown() then
		local sets = self.sets
		local w,h = 30, 30
		local newWidth = max((((30 + sets.spacing) * sets.columns) - sets.spacing) +(sets.padding*2), 8)
		local newHeight = max((((30 + sets.spacing) * sets.rows) - sets.spacing) +(sets.padding*2), 8)
		self:SetSize(newWidth, newHeight)

		local hori, vert, padhori, padvert
		
		if not sets.isRightToLeft then	
			hori = "Left"
			padhori = sets.padding
		else
			hori = "Right"
			padhori = -sets.padding
		end
		if not sets.isBottomToTop then
			vert = "Top"
			padvert = -sets.padding
		else
			vert = "Bottom"
			padvert = sets.padding
		end

		self.headerAura:ClearAllPoints()
		self.headerAura:SetPoint(vert..hori, self, vert..hori, padhori, padvert)
		local dir = "+"
		if (sets.direction == 1) or (sets.direction == true) then
			dir = "-"
		end

		local method = items[sets.method]
		self.headerAura:UpdateLayout(sets.spacing, sets.columns, sets.rows, sets.isRightToLeft, sets.isBottomToTop, method, dir)
		self:GetScript("OnEvent")()
		self:GetOrCreateOverlays()
		if 	self.HideBlizz then
			self:HideBlizz()
		end
	end
end

local function NewSlider(panel, name, min, max, arg)
        local slider =  panel:NewSlider({
            name = name,
            min = min,
            max = max,
            get = function(self) --Getter
                return panel.owner.sets[arg]
            end,
            set = function(self) --Setter
                local owner = panel.owner
                owner.sets[arg] = self:GetValue()
                owner:Layout()
                return owner.sets[arg]
            end,
        })
end

local function NewCheckButton(panel, name, arg)
        local c = panel:NewCheckButton{
            name = name,
            get = function() return panel.owner.sets[arg] end,
            set = function(self, enable)
                panel.owner.sets[arg] = self:GetChecked()
                panel.owner:Layout()
            end
        }
end

local function NewDropdown(panel, name, arg, items)
        local c = panel:NewDropdown{
		name = name,

		get = function()
			return panel.owner.sets[arg]
		end,

		set = function(_, value)
		    panel.owner.sets[arg] = value
		    panel.owner:Layout()
		end,

		items = items
	}
end

local function AddLayoutPanel(menu)
	local p = menu:NewPanel("Layout")
	NewCheckButton(p, "Flip Vertical", "isRightToLeft")
	NewCheckButton(p, "Flip Horizontal", "isBottomToTop")
	p:NewScaleSlider()
    NewSlider(p, "Padding", -13, 32, "padding")
    NewSlider(p, "Columns", 1, 20, "columns")
    NewSlider(p, "Rows", 1, 20, "rows")
    p:NewSpacingSlider()
	p:NewFadeSlider()
	p:NewOpacitySlider()

end

--viable unit check
local base = {
	"player",
	"target",
	"pet",
	"focus",
	"arena%d+",
	"boss%d+",
	"party%d+",
	"partypet%d+",
	"raid%d+",
	"raidpet%d+",
	"mouseover",
}

local max = {
	raid = 40,
	raidpet = 40,
	arena = 4,
	boss = 4,
	party = 4,
	partypet = 4,
}

local function IsViable(unit)
	local keep = unit
	unit = string.lower(unit)
	local count = 0
	if string.match (unit, "target" , 2) then
		_, count = string.gsub(string.sub (unit, 2, string.len(unit)), "target", "")
		if (unit == "player") and count > 0 then
			return nil
		else
			unit = string.gsub (unit, "target", "", count)
		end
	end	
	local can
	for test, match in pairs(base) do
		if string.match(string.gsub(match, "%d+", ""), string.gsub(unit, "%d+", "") ) then
			hasLimit = max[string.gsub( unit, "%d+", "" ) ]
			if hasLimit then
				local o = string.gsub(unit, string.gsub(unit, "%d+", ""),"")
				o = tonumber(o)
				if o and ( (o < 1) or (o>hasLimit) ) then
					return nil
				end
			end
			return string.lower(keep)
		end
	end
end

local hideBlizz
local function HideBlizz(panel)
	if not hideBlizz then
		NewCheckButton(panel, "Disable Default Buffs", "hideBlizz")
		hideBlizz = true
	end
end

local function AddAdvancedOptions(menu)
    local L = LibStub('AceLocale-3.0'):GetLocale('Dominos-Config')

    local panel = menu:NewPanel(L.Advanced)
	HideBlizz(panel)
	panel:NewClickThroughCheckbox()
	panel:NewShowInOverrideUICheckbox()
	panel:NewShowInPetBattleUICheckbox()
	panel.showStatesEditBox = panel:NewTextInput{
		name = L.ShowStates,
		multiline = true,
		width = 290,
		height = 64,
		get = function() return panel.owner:GetShowStates() end,
		set = function(_, value) panel.owner:SetShowStates(value) end
	}

	panel.TargetingEditBox = panel:NewTextInput{
		name = "Target",
		multiline = true,
		width = 290,
		height = 64,
		get = function() return panel.owner:GetTarget() or '' end,
		set = function(self, value)
            local text = value
            if IsViable(text) then
                panel.owner:SetTarget(text ~= '' and text or nil)
            else
                print("Invalid Unit: "..text)
            end
            panel.TargetingEditBox.editBox:ClearFocus()
        end
	}
end

local function AddSortPanel(self)
	local p = self:NewPanel("Sorting")

    NewCheckButton(p, "Reverse", "direction")
    NewDropdown(p, "Sort Method", "method", {
			{text = "Time", value = '1'},
			{text = "Index", value = '2'},
			{text = "Name", value = '3'},
		})

	return p
end

function temp:CreateMenu()
	local menu = Dominos:NewMenu(self.id)
	AddLayoutPanel(menu)
	AddSortPanel(menu)
	AddAdvancedOptions(menu)
	self.menu = menu
end
